@@ -10,7 +10,7 @@ import tempfile
 import configparser
 from subprocess import run, PIPE, DEVNULL
 
-version = "0.8.2"
+version = "0.9.0"
 
 
 class PassNode(urwid.AttrMap):
@@ -45,9 +45,11 @@ class PassList(urwid.ListBox):
     def __init__(self, body, root='', ui=None):
         self._ui = ui
         self.root = root
+        self._size = (1, 1)
         super().__init__(body)
 
     def mouse_event(self, size, event, button, col, row, focus):
+        self._size = size
         focus_offset = self.get_focus_offset_inset(size)[0]
 
         logging.debug("passlist mouse event: {} {} {} {} {} {} {} {}".format(
@@ -61,22 +63,23 @@ class PassList(urwid.ListBox):
                 if row == self.focus_position:
                     self.dir_navigate('down')
                 else:
-                    self.list_navigate(size, new_focus=row)
+                    self.list_navigate(new_focus=row)
             else:
                 if row == focus_offset:
                     self.dir_navigate('down')
                 else:
-                    self.list_navigate(size, new_focus=self.focus_position - focus_offset + row)
+                    self.list_navigate(new_focus=self.focus_position - focus_offset + row)
         elif button == 3:
             self.dir_navigate('up')
         elif button == 4:
-            self.list_navigate(size, -1)
+            self.list_navigate(-1)
         elif button == 5:
-            self.list_navigate(size, 1)
+            self.list_navigate(1)
         else:
             return super().mouse_event(size, event, button, col, row, focus)
 
     def keypress(self, size, key):
+        self._size = size
         logging.debug("passlist keypress: {} {}".format(key, size))
 
         list_navigation_offsets = {
@@ -99,7 +102,7 @@ class PassList(urwid.ListBox):
 
         action = config.keybindings.get(key)
         if action in list_navigation_offsets:
-            self.list_navigate(size, list_navigation_offsets[action])
+            self.list_navigate(list_navigation_offsets[action])
         elif action in dir_navigation_directions:
             self.dir_navigate(dir_navigation_directions[action])
         else:
@@ -123,8 +126,8 @@ class PassList(urwid.ListBox):
 
         self._ui.update_view()
 
-    def list_navigate(self, size, shift=0, new_focus=None):
-        offset = self.get_focus_offset_inset(size)[0]
+    def list_navigate(self, shift=0, new_focus=None):
+        offset = self.get_focus_offset_inset(self._size)[0]
 
         # either specify a shift offset, or an absolute position
         if new_focus is None:
@@ -135,10 +138,10 @@ class PassList(urwid.ListBox):
 
         # border check
         new_focus = min(max(new_focus, 0), len(self.body) - 1)
-        new_offset = min(max(new_offset, 0), size[1] - 1)
+        new_offset = min(max(new_offset, 0), self._size[1] - 1)
 
-        self.change_focus(size, new_focus, offset_inset=new_offset)
-        self._ui.update_preview()
+        self.change_focus(self._size, new_focus, offset_inset=new_offset)
+        self._ui.update_view()
 
     def insert(self, node):
         passnode = PassNode(node, self.root)
@@ -203,7 +206,9 @@ class UI(urwid.Frame):
         self._app_string = 'cPass'
         self._preview_shown = True
         self._edit_type = None
-        self._help_string = ' a:generate e:edit i:insert y:copy z:toggle'
+        self._search_pattern = None
+        self._search_direction = 1
+        self._help_string = ' a:generate e:edit i:insert y:copy z:toggle /:search'
 
         # widgets
         self.path_indicator = urwid.Text('', wrap='clip')
@@ -268,8 +273,12 @@ class UI(urwid.Frame):
         elif self._edit_type is not None:
             # pass through to edit widget (the focused widget)
             return super().keypress(size, key)
-        elif action == 'search':
-            self.focus_edit("search", '/')
+        elif action == 'search' or action == 'search_back':
+            self.focus_edit("search", '/' if action == 'search' else '?')
+            self._search_direction = 1 if action == 'search' else -1
+        elif action == 'search_next' or action == 'search_prev':
+            self.search_in_dir(self._search_pattern,
+                               1 if action == 'search_next' else -1)
         elif action == 'insert':
             self.focus_edit("insert", 'Enter password filename: ')
         elif action == 'generate':
@@ -308,8 +317,9 @@ class UI(urwid.Frame):
 
     def handle_input(self):
         if self._edit_type == "search":
-            # TODO: search
+            self._search_pattern = self.editbox.edit_text
             self.unfocus_edit()
+            self.search_in_dir(self._search_pattern, 1)
         elif self._edit_type == "generate":
             # TODO: can not accept "dir/node" format
             self.unfocus_edit()
@@ -432,6 +442,30 @@ class UI(urwid.Frame):
         else:
             self.message("Nothing copied", alert=True)
 
+    def search_in_dir(self, pattern, direction):
+        """ direction = 1 or -1 to specify the search direction """
+        if self._search_pattern is None:
+            self.message("No search pattern", alert=True)
+            return
+
+        # search from the next/previous, wrap if reach bottom/top
+        direction *= self._search_direction
+        start = self.listbox.focus_position
+        length = len(self.listbox.body)
+
+        # The math here is kind of magic. It's the result after simplification
+        if direction > 0:
+            start += 1
+        search_list = list(range(start, length)) + list(range(start))
+
+        for i in search_list[::direction]:
+            node = self.listbox.body[i].node
+            if node.find(pattern) != -1:
+                self.listbox.list_navigate(new_focus=i)
+                return
+
+        self.message("No matching", alert=True)
+
 
 class Pass:
     FALLBACK_PASS_DIR = os.path.join(os.getenv("HOME"), ".password_store")
@@ -524,6 +558,9 @@ class MyConfigParser(configparser.RawConfigParser):
             'cancel': ['esc'],
             'confirm': ['enter'],
             'search': ['/'],
+            'search_back': ['?'],
+            'search_next': ['n'],
+            'search_prev': ['N'],
             'insert': ['i'],
             'generate': ['a'],
             'edit': ['e'],
